/*
 * textbox.c
 *
 * 
 * Author: 
 *  Richard Hult <rhult@hem.passagen.se>
 * 
 *  http://www.dtek.chalmers.se/~d4hult/oregano/ 
 * 
 * Copyright (C) 1999,2000  Richard Hult 
 * 
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of the 
 * License, or (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <config.h>
#include <gnome.h>
#include <math.h>
#include "item-data.h"
#include "textbox.h"
#include "clipboard.h"

#define TEXTBOX_DEFAULT_FONT "-*-helvetica-medium-r-*-*-*-100-*-*-*-*-*-*"

static void      textbox_class_init (TextboxClass *klass);
static void      textbox_init (Textbox *textbox);
static void      textbox_copy (ItemData *dest, ItemData *src);
static ItemData *textbox_clone (ItemData *src);
static void      textbox_rotate (ItemData *data, int angle, SheetPos *center);
static void      textbox_flip (ItemData *data, gboolean horizontal, SheetPos *center);

enum {
	TEXT_CHANGED,
	FONT_CHANGED,
	LAST_SIGNAL
};

static ItemDataClass *parent_class = NULL;
static guint textbox_signals[LAST_SIGNAL] = { 0 };

struct _TextboxPriv { 
	char *text;
	char *font; 
};

guint
textbox_get_type (void)
{
	static guint textbox_type = 0;

	if (!textbox_type) {
		static const GtkTypeInfo textbox_info = {
			"Textbox",
			sizeof (Textbox),
			sizeof (TextboxClass),
			(GtkClassInitFunc) textbox_class_init,
			(GtkObjectInitFunc) textbox_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc) NULL,
		};
	
		textbox_type = gtk_type_unique (TYPE_ITEM_DATA, &textbox_info);
	}

	return textbox_type;
}

static void
textbox_destroy (GtkObject *object)
{
	Textbox *textbox = TEXTBOX (object);
	TextboxPriv *priv = textbox->priv;

	g_free (priv);
	
	GTK_OBJECT_CLASS (parent_class)->destroy (object);
}

static void
textbox_class_init (TextboxClass *klass)
{
	GtkObjectClass *object_class;
	ItemDataClass *item_data_class;

	parent_class = gtk_type_class (TYPE_ITEM_DATA);
	item_data_class = (ItemDataClass *) klass;
	object_class = (GtkObjectClass *) klass;

	textbox_signals[TEXT_CHANGED] = 
		gtk_signal_new ("text_changed",
				GTK_RUN_FIRST,
				object_class->type,
				0,
				gtk_marshal_NONE__STRING,
				GTK_TYPE_NONE, 1, GTK_TYPE_STRING);

	textbox_signals[FONT_CHANGED] = 
		gtk_signal_new ("font_changed",
				GTK_RUN_FIRST,
				object_class->type,
				0,
				gtk_marshal_NONE__STRING,
				GTK_TYPE_NONE, 1, GTK_TYPE_STRING);
	
	gtk_object_class_add_signals (object_class, textbox_signals, LAST_SIGNAL);

	object_class->destroy = textbox_destroy;

	item_data_class->clone = textbox_clone;
	item_data_class->copy = textbox_copy;
	item_data_class->rotate = textbox_rotate;
	item_data_class->flip = textbox_flip;
	item_data_class->unreg = NULL;
	item_data_class->reg = NULL;
}

static void
textbox_init (Textbox *textbox)
{
	TextboxPriv *priv = g_new0 (TextboxPriv, 1);
	textbox->priv = priv;
}

Textbox *
textbox_new (char *font)
{
	Textbox *textbox;

	textbox = gtk_type_new (textbox_get_type ());

	if (font == NULL)
		textbox->priv->font = g_strdup (TEXTBOX_DEFAULT_FONT);
	else
		textbox->priv->font = g_strdup (font);

	return textbox;
}

static ItemData *
textbox_clone (ItemData *src)
{
	Textbox *src_textbox, *new_textbox;
	ItemDataClass *id_class;
	
	g_return_val_if_fail (src != NULL, NULL);
	g_return_val_if_fail (IS_TEXTBOX (src), NULL);

	id_class = ITEM_DATA_CLASS (GTK_OBJECT(src)->klass);
	if (id_class->copy == NULL)
		return NULL;

	src_textbox = TEXTBOX (src);
	new_textbox = gtk_type_new (TYPE_TEXTBOX);
	id_class->copy (ITEM_DATA (new_textbox), src);

	return ITEM_DATA (new_textbox);
}

static void
textbox_copy (ItemData *dest, ItemData *src)
{
	Textbox *dest_textbox, *src_textbox;
	
	g_return_if_fail (dest != NULL);
	g_return_if_fail (IS_TEXTBOX (dest));
	g_return_if_fail (src != NULL);
	g_return_if_fail (IS_TEXTBOX (src));

	if (parent_class->copy != NULL)
		parent_class->copy (dest, src);

	dest_textbox = TEXTBOX (dest);
	src_textbox = TEXTBOX (src);

	dest_textbox->priv->text = src_textbox->priv->text;
	dest_textbox->priv->font = src_textbox->priv->font;
}

static void
textbox_rotate (ItemData *data, int angle, SheetPos *center)
{
	double affine[6];
	ArtPoint src, dst;
	Textbox *textbox;
	TextboxPriv *priv;
	SheetPos b1, b2;
	SheetPos textbox_center, delta;

	g_return_if_fail (data != NULL);
	g_return_if_fail (IS_TEXTBOX (data));

	if (angle == 0)
		return;

	textbox = TEXTBOX (data);

	if (center) {
		item_data_get_absolute_bbox (ITEM_DATA (textbox), &b1, &b2);
		textbox_center.x = b1.x + (b2.x - b1.x) / 2;
		textbox_center.y = b1.y + (b2.y - b1.y) / 2;
	}

	priv = textbox->priv;

	art_affine_rotate (affine, angle);

	/*
	 * Let the views (canvas items) know about the rotation.
	 */
	gtk_signal_emit_by_name (GTK_OBJECT (textbox), "rotated", angle);

	if (center) {
		SheetPos textbox_pos;

		item_data_get_pos (ITEM_DATA (textbox), &textbox_pos);

		src.x = textbox_center.x - center->x;
		src.y = textbox_center.y - center->y;
		art_affine_point (&dst, &src, affine);
		
		delta.x = -src.x + dst.x;
		delta.y = -src.y + dst.y;
		
		item_data_move (ITEM_DATA (textbox), &delta);		
	}
}

static void
textbox_flip (ItemData *data, gboolean horizontal, SheetPos *center)
{
	double affine[6];
	ArtPoint src, dst;
	Textbox *textbox;
	TextboxPriv *priv;
	SheetPos b1, b2;
	SheetPos textbox_center, delta;

	g_return_if_fail (data != NULL);
	g_return_if_fail (IS_TEXTBOX (data));

	textbox = TEXTBOX (data);

	if (center) {
		item_data_get_absolute_bbox (ITEM_DATA (textbox), &b1, &b2);
		textbox_center.x = b1.x + (b2.x - b1.x) / 2;
		textbox_center.y = b1.y + (b2.y - b1.y) / 2;
	}

	priv = textbox->priv;

	if (horizontal)
		art_affine_scale (affine, -1, 1);
	else
		art_affine_scale (affine, 1, -1);

	/*
	 * Let the views (canvas items) know about the rotation.
	 */
	gtk_signal_emit_by_name (GTK_OBJECT (textbox), "flipped", horizontal);

	if (center) {
		SheetPos textbox_pos;

		item_data_get_pos (ITEM_DATA (textbox), &textbox_pos);

		src.x = textbox_center.x - center->x;
		src.y = textbox_center.y - center->y;
		art_affine_point (&dst, &src, affine);
		
		delta.x = -src.x + dst.x;
		delta.y = -src.y + dst.y;
		
		item_data_move (ITEM_DATA (textbox), &delta);		
	}
}

/* static */
void
textbox_update_bbox (Textbox *textbox)
{
	GdkFont *font;
	int width;
	int rbearing;
	int lbearing;
	int ascent, descent;
	SheetPos b1, b2;
	TextboxPriv *priv;

	priv = textbox->priv;

	font = gdk_font_load (priv->font);
	gdk_string_extents (font,
			    priv->text,
			    &lbearing,
			    &rbearing,
			    &width,
			    &ascent,
			    &descent);
	gdk_font_unref (font);
	
	b1.x = 0.0;
	b1.y = 0.0 - font->ascent;
	b2.x = 0.0 + rbearing;
	b2.y = 0.0 + font->descent; 
		
	item_data_set_relative_bbox (ITEM_DATA (textbox), &b1, &b2);
}

void
textbox_set_text (Textbox *textbox, char *text)
{
	g_return_if_fail (textbox != NULL);
	g_return_if_fail (IS_TEXTBOX (textbox));

	g_free (textbox->priv->text);
	textbox->priv->text = g_strdup (text);

	textbox_update_bbox (textbox);

	gtk_signal_emit (GTK_OBJECT (textbox), textbox_signals[TEXT_CHANGED], text);
}

char *
textbox_get_text (Textbox *textbox)
{
	g_return_val_if_fail (textbox != NULL, NULL);
	g_return_val_if_fail (IS_TEXTBOX (textbox), NULL);

	return textbox->priv->text;
}

void
textbox_set_font (Textbox *textbox, char *font)
{
	g_return_if_fail (textbox != NULL);
	g_return_if_fail (IS_TEXTBOX (textbox));

	g_free (textbox->priv->font);
	if (font == NULL)
		textbox->priv->font = g_strdup (TEXTBOX_DEFAULT_FONT);
	else
		textbox->priv->font = g_strdup (font);

	textbox_update_bbox (textbox);

	gtk_signal_emit (GTK_OBJECT (textbox), textbox_signals[FONT_CHANGED], textbox->priv->font);
}

char *
textbox_get_font (Textbox *textbox)
{
	g_return_val_if_fail (textbox != NULL, NULL);
	g_return_val_if_fail (IS_TEXTBOX (textbox), NULL);

	return textbox->priv->font;
}



